(function() {
  'use strict';

  /**
   *
   * Metavera API service implementation
   *
   * Naming conventions:
   *  - fetch* methods returns promise which is going to be resolved after actual data arrived
   *  - get* methods should immediately return the value
   *
   * @param $http
   * @constructor
   *
   */
  function MetaVeraService($http, $q, x2js, moment) {

    //Checking are we in devmode or not
    this.getQueryVariable = function(){

      var vars = [], hash;
      var hashes = window.location.href.slice(window.location.href.indexOf('?') + 1).split('&');

      for(var i = 0; i < hashes.length; i++){
        hash = hashes[i].split('=');
        vars.push(hash[0]);
        vars[hash[0]] = hash[1];
      }

      return vars;

    };

    var serviceUrl = 'https://reserve.streetcar.ru/webservices/index.php/WSUser/WSRest';

    if ( this.getQueryVariable() == 'dev' ) {
      serviceUrl = 'http://localhost:3030';
    }

    this.clientTime = undefined;
    this.hash = undefined;
    this.billcode = undefined;

    this.tripTimeResolution = 15;

    /**
     *
     * Calculates sha1 hash for auth purposes for MetaVera api.
     * Refer to Mobile API quickstart guide.
     *
     * @param method
     * @param clientTime
     * @returns {string}
     */
    this.calculateHash = function(method, clientTime) {
      //$hash = sha1(sha1($password).$clientTime.$method);
      var SHA1 = new Hashes.SHA1;
      return SHA1.hex(SHA1.hex(this.password) + clientTime + method);
    };

    this.getTime = function() {
      return Math.round(Date.now()/1000);
    };
    /**
     *
     * Performs request to MetaVera API with given method and parameters
     * Returns promise which could be used later
     *
     * @param method
     * @param params
     * @returns {HttpPromise}
     */

    this.doRequest = function (method, params, cache) {

      // saving context
      var _self = this;

      // Fetching current time
      var time = this.getTime();

      // Caching is not implemented yet
      var cacheVal = cache || false;

      // Building request dictionary
      var req = {
        params: {
          action: method,
          user: this.user,
          hash: this.calculateHash(method, time),
          time: time,
          billcode: 'service',
          format: 'json'
        },

        transformResponse:function (data) {
          // convert the data to JSON and provide
          // it to the success function below
          var json = x2js.xml_str2json(data);

          return json;
        }
      };

      angular.extend(req.params, params);

      return $http.get(serviceUrl, req).then(function(data) {
        // Check if normal response
        if ( data.data && data.data.methodResponse ) {
          // Check if there is any error
          if ( data.data.methodResponse.fault ) {
            return $q.reject(_self.errorToDict(data.data.methodResponse));
          }

          return data.data.methodResponse;
        }
        return data.data;
      });
    };

    /**
     * Converts complex json error structure to flat one
     *
     * @param struct
     * @returns ErrorDict
     */
    this.errorToDict = function(e) {
      try {
        return {
          'code': e.fault.value.struct.member[0]['value']['int'],
          'string': e.fault.value.struct.member[1]['value']['string']
        }
      } catch(e) {
        return {'code': "EE", 'string': "Error while unpacking error"};
      }

    };

    /**
     *
     * Performs auth using 'isLoggedIn' service on MetaVera service
     * Additionally it saves login and password for later use
     *
     * @param login
     * @param password
     * @returns {HttpPromise}
     */
    this.login = function(login, password) {

      // Save user and password
      this.user = login;
      this.password = password;

      // Make request call
      return this.doRequest('isLoggedIn').then(function(response) {
        return response;
      });
    };


    /**
     *
     * Fetch Trip Time Resolution parameter from MetaVera API
     *
     * @returns {HttpPromise}
     */
    this.fetchTripTimeResolution = function() {
      var self = this;
      return this.doRequest('tripTimeResolution').then(function(data) {
        self.tripTimeResolution = Number(data) || 15;
      });
    };

    /**
     *
     * Search available cars using 'resultsFromStackFilter' MetaVera API call
     *
     * @returns {HttpPromise}
     */
    this.searchAvailableCars = function(timestep, lat, long) {

      function roundTo(num, div){
        return Math.floor(num/div)*div;
      }

      var params = {
        'aStackFilter[startTime]': roundTo(this.getTime() + 900, 60 * this.tripTimeResolution),
        'aStackFilter[endTime]': roundTo((this.getTime() + 60 * 60 * timestep * 1) + 900, 60 * this.tripTimeResolution),
        includeStack: 'true' };

      if ( lat ) { params['aStackFilter[latitude]']=lat; }
      if ( long ) { params['aStackFilter[longitude]'] = long; }

      return this.doRequest('resultsFromStackFilter', params);
    };

    this.fetchCurrentAndFutureReservations = function() {
      return this.doRequest('getCurrentAndFutureReservations').then(function (data) {
        if (!('DBEntityReservation' in data.WSGetCurrentAndFutureReservationsResult.futureReservations)) {
          data.WSGetCurrentAndFutureReservationsResult.futureReservations.DBEntityReservation = [];
          return data;
        }
        if (!( 'length' in data.WSGetCurrentAndFutureReservationsResult.futureReservations.DBEntityReservation)) {
          data.WSGetCurrentAndFutureReservationsResult.futureReservations.DBEntityReservation = [data.WSGetCurrentAndFutureReservationsResult.futureReservations.DBEntityReservation];
        }
        return data;
      });
    };

    this.driverName = function() {
      return this.doRequest('getDriverName');
    };

    this.getConfiguration = function() {
      return this.doRequest('getConfiguration');
    };

    this.Locale = function() {
      return this.doRequest('getDriverLanguageLocale');
    };

    this.fetchcurrentReservation = function() {
      return this.doRequest('currentReservation');
    };

    this.fetchPastReservations = function() {
      return this.doRequest('pastReservations');
    };

    this.fetchFutureReservations = function () {
      return this.doRequest('futureReservations');
    };

    this.fetchPastTrips = function () {
      return this.doRequest('pastTrips');
    };

    this.fetchMessages = function() {
      return this.doRequest('getDriverMessages');
    };

    this.fetchReservationById = function(id) {
      return this.doRequest('reservationWithId', {'reservationId': id}, true);
    };

    this.virtualSwipeIn = function () {
      return this.doRequest('virtualSwipeIn');
    };

    this.virtualSwipeOut = function () {
      return this.doRequest('virtualSwipeOut');
    };

    this.startTripById = function (id) {
      return  this.doRequest('startTripById', {'reservationId': id}, true);
    };

    this.endTripById = function (id) {
      return  this.doRequest('endTripById', {'reservationId': id, 'endOdo': ''}, true);
    };

    this.cancelReservationWithId = function (id) {
      return  this.doRequest('cancelReservationWithId', {'reservationId': id}, true);
    };

    this.extendReservationWithId = function (id, endTime) {
      return  this.doRequest('extendReservationWithId', {'reservationId': id, 'endTime': endTime}, true);
    };

    this.editReservationTimes = function (id, startTime, endTime) {
      return  this.doRequest('editReservationTimes', {
        'reservationId': id,
        startTime: startTime,
        'endTime': endTime
      }, true);
    };

    this.availabilityForStack = function (stackId, startTime, endTime) {
      return  this.doRequest('availabilityForStack', {
        'stackId': stackId,
        'startTime': startTime,
        'endTime': endTime
      }, true);
    };

    this.vehicleById = function (vehicleId) {
      return  this.doRequest('vehicleById', {'vehicleId': vehicleId}, true);
    };

    this.allVehicleTypes = function () {
      return  this.doRequest('allVehicleTypes').then(function(vehicles){
        return vehicles.DBEntityVehicleType;
      });
    };

    this.tripEstimate = function (stackId, startTime, endTime) {
      return  this.doRequest('tripEstimate', {
        'stackId': stackId,
        'startTime': startTime,
        'endTime': endTime
      }, true);
    };

    this.makeReservationGetId = function (stackId, startTime, endTime) {
      return  this.doRequest('makeReservationGetId', {
        'stackId': stackId,
        'startTime': startTime,
        'endTime': endTime,
        'memo': ''}, true);
    };

    /**
     * Perform virtual swipe-in using virtual card on given reservation.
     * Should wait for sentack state before it become fully operational
     *
     * @param id
     * @returns {HttpPromise}
     */
    this.virtualSwipeInOnReservation = function(id) {
      return this.doRequest('virtualSwipeInOnReservation', {'reservationID': id});
    };

    /**
     * Perform virtual swipe-out using virtual card on given reservation.
     * Should answer to questions on the terminal before it could be called
     *
     * @param id
     * @returns {HttpPromise}
     */
    this.virtualSwipeOutOnReservation = function (id) {
      return this.doRequest('virtualSwipeOutOnReservation', {'reservationID': id});
    };


    /**
     * Fetched swipe status
     *
     * @param id
     * @returns {HttpPromise}
     */
    this.getSwipeInStatus = function(id) {
      return this.doRequest('getVirtualSwipeStatus', {
        'swipeId': id
      });
    };

    this.getReservationMessages = function(id){
      return this.doRequest('getReservationMessages', { 'reservationID': id });
    };
  }

  /**
   *
   * Basic angular wrapper for x2js
   *
   * @returns {X2JS}
   */
  function x2js() {
    return new X2JS({
      arrayAccessFormPaths: [
//        'data.WSGetCurrentAndFutureReservationsResult.futureReservations.DBEntityReservation'
      ],
      emptyNodeForm: 'string'

    });
  }

  angular
    .module('app.services.metavera', [])
    .factory("x2js", x2js)
    .service('MetaVera', MetaVeraService);

}());
