(function() {
    'use strict';

    function RegistrationService($http, $q, $localStorage) {
        var regUrl = 'https://www.streetcar.ru/api/v1/register';
        var sessionId;
        var applicantId;

        if ( $localStorage.getObject('sessionId') ) {
            console.log("Initialize session");
            sessionId = $localStorage.getObject('sessionId').sId;
        }

        this.takePlans = function(){
            return $http.get('https://streetcar.ru/api/v1/register/plans').then(function (data) {
                return data;
            });
        };

        /**
         * Step 1
         * Post data of new user to StreetCars registration API(SCREGAPI) with given parameters
         * Returns promise with session id
         * or returns us an ErrorDict with error code and message
         *
         * @param params
         * @returns {HttpPromise}
         */
        this.registerUser = function(params) {
            var _self = this;
            return $http.post(regUrl, params).then(function(data){
                if ( data && ( data.data.status == 'success' ) ) {
                    _self.saveSId(data.data.sid);
                    return data;
                } else if ( data && data.data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            }, function(data) {
                // There is a registration in progress
                if (data.data.status == 'error' && data.data.code == "2") {
                    _self.saveSId(data.data.sid);
                    return data;
                }
                return $q.reject(_self.errorToDict(data));
            });
        };

        this.saveSId = function(sId){
            sessionId = sId;
            return $localStorage.setObject('sessionId', {sId: sId});
        };

        /**
         * Step 2 (passport)
         * Post photo of new user passport to SCREGAPI
         * Returns promise with status
         *
         * @param photo
         * @returns {HttpPromise}
         */
        this.sendPassport = function(photo) {
            var _self = this;
            var fd = new FormData();
            fd.append('passportFile',photo, "passport.png");

            return $http.post([regUrl, sessionId, 'passport'].join('/'), fd, {
                transformRequest: angular.identity,
                headers: {'Content-Type': undefined}
            }).then(function(data){
                if ( data && data.status == 'success' ) {
                    return data;
                } else if ( data && data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            });
        };

        /**
         * Step 3 (license)
         * Post photo of new user driver license to SCREGAPI
         * Returns promise with status
         *
         * @param photo
         * @returns {HttpPromise}
         */
        this.sendLicense = function(photo) {
            var _self = this;
            var fd = new FormData();
            fd.append('licenseFile',photo, "license.png");
            return $http.post([regUrl, sessionId, 'license'].join('/'), fd, {
                transformRequest: angular.identity,
                headers: {'Content-Type': undefined}
            }).then(function(data){
                if ( data && data.status == 'success' ) {
                    return data;
                } else if ( data && data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            });
        };

        /**
         * Step 4 (agreement, personal)
         * Get the user agreement from SCREGAPI
         * Returns success response with user agreement
         *
         * @returns {HTML}
         */
        this.getAgreement = function() {
            var _self = this;
            return $http.get([regUrl, sessionId, 'agreement'].join('/')).then(function(data){
                if ( data.data ) {
                    return data.data;
                } else if ( data && data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            });
        };

        this.getLicense = function(link) {
            var _self = this;
            return $http.get([regUrl, link].join('/'), {cache: true}).then(function(data){
                if ( data.data ) {
                    return data.data;
                } else if ( data && data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            });
        };

        /**
         * Step 5 (smscode, smscoderetry, smscodecheck)
         * Verify registration of new user via SMS code
         * Returns promise with status
         *
         * @returns {HttpPromise}
         */
        this.smsSend = function() {
            var _self = this;
            return $http.get([regUrl, sessionId, 'smscode'].join('/')).then(function(data){
                if ( data.data && data.data.status == 'success' ) {
                    return data;
                } else if ( data.data && data.data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            });
        };

        this.smsAgain = function() {
            var _self = this;
            return $http.get([regUrl, sessionId, 'smscoderetry'].join('/')).then(function(data){
                if ( data && data.status == 'success' ) {
                    return data;
                } else if ( data && data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            });
        };

        this.smsVerify = function(code) {
            var _self = this;
            return $http.post([regUrl, sessionId, 'smscodecheck'].join('/'), { telSmsCode: code }).then(function(data){
                if ( data && data.status == 'success' ) {
                    return data;
                } else if ( data && data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(data));
                }
            });
        };



        /**
         * Step 6 (card)
         * Put data of user eCard to SCREGAPI
         * Returns promise with status and applicantID
         *
         * @param params
         * @returns {HttpPromise}
         */
        this.saveCardData = function(params) {
            var _self = this;
            return $http.put([regUrl, sessionId, 'card'].join('/'), params ).then(function(resp){
                if ( resp && resp.data.status == 'success' ) {
                    applicantId = resp.data.applicantID;
                    return resp.data;
                } else if ( resp && resp.data.status == 'error' ) {
                    // Check if there is any error
                    return $q.reject(_self.errorToDict(resp));
                }
                return resp.data;
            });
        };

        /**
         * Converts complex json error structure to flat one
         *
         * @returns ErrorDict
         */
        this.errorToDict = function(e) {
            try {
                return {
                    'code': e.data.code,
                    'message': e.data.message
                }
            } catch(e) {
                return {'code': "E", 'string': "Error while unpacking error"};
            }

        };

        this.getAppId = function() {
            return applicantId;
        }
    }

    angular
        .module('app.services.registration', [])
        .service('RegistrationService', RegistrationService);

}());
