(function () {
  'use strict';

  /**
   * User service implmementation
   *
   * @NgAnnotate
   */
  function UserService($rootScope, MetaVera, $localStorage, $ionicHistory) {
    var sessionData = {};
    var pastData = {};
    var driver = {};
    var vehicles = {};
    var sessionDataPromise;
    var pastDataPromise;
    var driverPromise;
    var vehiclesPromise;
    var self = this;

    /**
     * Fetch all required data for app operation
     *
     */
    this.fetchSessionData = function() {
      sessionDataPromise = MetaVera.fetchCurrentAndFutureReservations().then(function(r) {
        sessionData.reservations = r;
        return r;
      });
      return sessionDataPromise;
    };

    this.fetchPastData = function() {
      pastDataPromise = MetaVera.fetchPastReservations().then(function(r) {
        pastData.reservations = r;
        return r;
      });
      return pastDataPromise;
    };

    this.fetchDriver = function() {
      driverPromise = MetaVera.driverName().then(function(r) {
        return r;
      });
      return driverPromise;
    };

    this.fetchVehicles = function() {
      vehiclesPromise = MetaVera.allVehicleTypes().then(function(r) {
        return r;
      });
      return vehiclesPromise;
    };

    this.resolveSessionData = function() {
      if ( sessionDataPromise ) {
        return sessionDataPromise;
      }
      return this.fetchSessionData();
    };

    this.resolvePastData = function() {
      if ( pastDataPromise ) {
        return pastDataPromise;
      }
      return this.fetchPastData();
    };

    this.resolveDriverData = function() {
      if ( driverPromise ) {
        return driverPromise;
      }
      return this.fetchDriver();
    };

    this.resolveVehiclesData = function() {
      if ( vehiclesPromise ) {
        return vehiclesPromise;
      }
      return this.fetchVehicles();
    };

    /**
     * Returns fetched session data
     *
     * @returns {SessionData}
     */
    this.getSessionData = function() {
      return sessionData;
    };

    this.getPastData = function() {
      return pastData;
    };

    this.getDriverData = function() {
      return driver;
    };

    this.getVehiclesData = function() {
      return vehicles;
    };

    /**
     * Initializes user on service start
     *
     */
    this.initUser = function() {
      var user = $localStorage.getObject('user');

      // Check if user has been saved

      if ( user ) {
        return this.login( user.user, user.password).then(function() {
          self.fetchSessionData();
          self.fetchPastData();
          self.fetchDriver();
          self.fetchVehicles();
        });
      }
    };

    /**
     * Performs login check with MetaVera API
     *
     * @param username
     * @param password
     * @returns {Promise}
     */
    this.login = function(username, password) {
      return MetaVera.login(username,  password).then(function(r) {
        $localStorage.setObject('user', { user: username,  password: password });
        self.fetchSessionData();
        self.fetchPastData();
        self.fetchDriver();
        self.fetchVehicles();

        $ionicHistory.clearCache();
        $ionicHistory.clearHistory();
        $rootScope.$broadcast('auth:logged');
        return r;
      });
    };

    this.logout = function() {
      $localStorage.removeObject('user');
      $localStorage.removeObject('lang');
      MetaVera.user = undefined;
      MetaVera.password = undefined;
    };

    this.initUser();
  }


  /**
   * Drop-in service for using with localstorage
   *
   * @param $window
   * @returns {{set: Function, get: Function, setObject: Function, getObject: Function}}
   */
  function localStorageService($window) {
    return {
      set: function (key, value) {
        $window.localStorage[key] = value;
      },
      get: function (key, defaultValue) {
        return $window.localStorage[key] || defaultValue;
      },
      removeObject: function(key) {
        delete $window.localStorage[key];
      },
      setObject: function (key, value) {
        $window.localStorage[key] = JSON.stringify(value);
      },
      getObject: function (key) {
        try {
          return JSON.parse($window.localStorage[key] || '{}');
        } catch(E) {

        }
        return null;
      }
    }
  }

  angular
    .module('app.services.user', ['app.services.metavera', 'angular-lodash'])
    .service('User', UserService)
    .factory('$localStorage', localStorageService);
}());
